#!/bin/python
# -*- coding: utf-8 -*-
"""
build_help.py

Author: GrimAndGreedy
License: MIT
"""

from keys import picker_keys
from listpick.listpick_app import Picker, start_curses, close_curses

import curses

def build_help_rows(keys_dict: dict) -> list[list[str]]:
    """ Build help rows based on the keys_dict. """
    ## Key names
    special_keys = {

            27: "Escape",
            353: "Shift+Tab",
            curses.KEY_END: "END",
            curses.KEY_HOME: "HOME",
            curses.KEY_PPAGE: "Page Up",
            curses.KEY_NPAGE: "Page Down",
            curses.KEY_UP: "ArrowUp",
            curses.KEY_DOWN: "ArrowDown",
            curses.KEY_RIGHT: "ArrowRight",
            curses.KEY_LEFT: "ArrowLeft",
            ord(' '): "Space",
            curses.KEY_ENTER: "RETURN",
            ord('\n'): "\n",
            curses.KEY_DC: "Delete",
    }

    # Ctrl + [a-z]
    for i in range(26):
        special_keys[i+1] = f"Ctrl+{chr(ord('a')+i)}"

    # F1-F12
    for i in range(12):
        special_keys[curses.KEY_F1+i] = f"F{i+1}"

    ## Key descriptions
    help_descriptions = {
        "refresh":                          "Refresh the screen.",
        "help":                             "Open help.",
        "exit":                             "Exit picker instance.",
        "full_exit":                        "Immediate exit to terminal.",
        "move_column_left":                 "Move column left.",
        "move_column_right":                "Move column right.",
        "cursor_down":                      "Cursor down.",
        "cursor_up":                        "Cursor up.",
        "half_page_up":                     "Half page up.",
        "half_page_down":                   "Half page down.",
        "page_up":                          "Page up.",
        "page_down":                        "Page down.",
        "cursor_bottom":                    "Send cursor to bottom of list.",
        "cursor_top":                       "Send cursor to top of list.",
        "five_up":                          "Five up.",
        "five_down":                        "Five down.",
        "toggle_select":                    "Toggle selection.",
        "select_all":                       "Select all.",
        "select_none":                      "Select none.",
        "visual_selection_toggle":          "Toggle visual selection.",
        "visual_deselection_toggle":        "Toggle visual deselection.",
        "enter":                            "Accept selections.",
        "redraw_screen":                    "Redraw screen.",
        "cycle_sort_method":                "Cycle through sort methods.",
        "cycle_sort_method_reverse":        "Cycle through sort methods (reverse)",
        "cycle_sort_order":                 "Toggle sort order.",
        "delete":                           "Delete dialaogue.",
        "decrease_lines_per_page":          "Decrease lines per page.",
        "increase_lines_per_page":          "Increase lines per page.",
        "increase_column_width":            "Increase column width.",
        "decrease_column_width":            "Decrease column width.",
        "filter_input":                     "Filter rows.",
        "search_input":                     "Search.",
        "settings_input":                   "Settings input.",
        "settings_options":                 "Settings options dialogue.",
        "continue_search_forward":          "Continue search forwards.",
        "continue_search_backward":         "Continue search backwards.",
        "cancel":                           "Cancel; escape.",
        "opts_input":                       "Options input.",
        "opts_select":                      "Options select dialogue.",
        "mode_next":                        "Cycle through modes forwards.",
        "mode_prev":                        "Cycle through modes backwards.",
        "pipe_input":                       "Pipe selected cells from selected rows.",
        "reset_opts":                       "Reset options.",
        "col_select":                       "Select column.",
        "col_select_next":                  "Select next column.",
        "col_select_prev":                  "Select previous column.",
        "col_hide":                         "Hide column.",
        "edit":                             "Edit cell.",
        "edit_picker":                      "Edit cell from options dialogue.",
        "edit_ipython":                     "Edit current data with ipython.",
        "copy":                             "Copy selections.",
        "save":                             "Save selections.",
        "load":                             "Load from file.",
        "open":                             "Open from file.",
        "toggle_footer":                    "Toggle footer.",
        "notification_toggle":              "Toggle empty notification.",
        "redo":                             "Redo.",
        "undo":                             "Undo.",
        "scroll_right":                     "Scroll right.",
        "scroll_left":                      "Scroll left.",
        "scroll_far_right":                 "Scroll to the end of the column set.",
        "scroll_far_left":                  "Scroll to the left home.",
    }

    # [[key_name, key_function_description], ...]
    items = []
    for val, keys in keys_dict.items():
        row = [[chr(int(key)) if key not in special_keys else special_keys[key] for key in keys], help_descriptions[val]]
        items.append(row)

    return items

items = build_help_rows(picker_keys)
stdscr = start_curses()
x = Picker(
        stdscr,
        items=items
        )
x.run()

close_curses(stdscr)



#!/bin/python
# -*- coding: utf-8 -*-
"""
input_field.py
Function to display an input field within a curses window.

Author: GrimAndGreedy
License: MIT
"""

import curses
from typing import Tuple, Optional, Callable

def input_field(
        stdscr: curses.window,
        usrtxt:str="",
        field_name:str="Input",
        x:Callable=lambda:0,
        y:Callable=lambda:0,
        colours_start:int=0,
        literal:bool=False,
        max_length:Callable = lambda: 1000,
        registers={},
        refresh_screen_function:Optional[Callable]=None,
        cursor: int = 0,
) -> Tuple[str, bool]:
    """
    Display input field at x,y for the user to enter text.

    ---Arguments
        stdscr: curses screen
        usrtxt (str): text to be edited by the user
        field_name (str): The text to be displayed at the start of the text input
        x (int): prompt begins at (x,y) in the screen given
        y (Callable): prompt begins at (x,y) in the screen given
        colours_start (int): where to start when initialising the colour pairs with curses.
        literal: whether to display the repr() of the string; e.g., if we want to display escape sequences literally
        max_length (callable): function that returns the length of input field


    ---Returns
        usrtxt, return_code
        usrtxt: the text inputted by the user
        return_code: 
                        0: user hit escape
                        1: user hit return
    """
    while True:

        h, w = stdscr.getmaxyx()

        if refresh_screen_function != None:
            refresh_screen_function()
        field_end = min(w-3, max_length())
        field_y = min(h-1, y())
        field_x = min(h-1, x())

        # Clear background to end of row
        stdscr.addstr(field_y, x(), " "*(field_end-x()), curses.color_pair(colours_start+20))
        stdscr.refresh()
        # Display the field name and current text
        field_length = 0

        if literal:
            stdscr.addstr(field_y, x(), f"{field_name}: {repr(usrtxt)}   "[:field_end], curses.color_pair(colours_start+13) | curses.A_BOLD)
            field_length=len(f"{field_name}: {repr(usrtxt)}   ")
        else:
            stdscr.addstr(field_y, x(), f" {field_name}: {usrtxt}   "[:field_end], curses.color_pair(colours_start+13) | curses.A_BOLD)
            field_length=len(f" {field_name}: {usrtxt}   ")

        visible_cursor_x = x()+len(usrtxt)-cursor+len(f" {field_name}: ")
        if literal:
            visible_cursor_x = x()+len(repr(usrtxt))-cursor+len(f" {field_name}: ")-2

        # if key == curses.KEY_RESIZE:  # Terminal resize signal

        # Display cursor if the field fits onto the screen
        if field_length + 1 < field_end:
            if not literal:
                if usrtxt and cursor != 0:
                    stdscr.addstr(field_y, visible_cursor_x, f"{usrtxt[-(cursor)]}", curses.color_pair(colours_start+13) | curses.A_REVERSE | curses.A_BOLD)
                else:
                    stdscr.addstr(field_y, visible_cursor_x, f" ", curses.color_pair(colours_start+13) | curses.A_REVERSE | curses.A_BOLD)
            elif literal:
                stdscr.addstr(field_y, visible_cursor_x, f"{repr(usrtxt)[-(cursor+1)]}", curses.color_pair(colours_start+13) | curses.A_REVERSE | curses.A_BOLD)

        key = stdscr.getch()

        if key == 27:                                                           # ESC key
            return "", False
        elif key == 3:                                                           # ESC key
            stdscr.keypad(False)
            curses.nocbreak()
            curses.noraw()
            curses.echo()
            curses.endwin()
            exit()
        elif key == 10:                                                         # Enter/return key
            return usrtxt, True
            # selected_indices = print_selected_indices()
            # if not selected_indices:
            #     selected_indices = [indexed_items[cursor_pos][0]]
            # full_values = [format_row_full(items[i], hidden_columns) for i in selected_indices]  # Use format_row_full for full data
            # # selected_data = [format_full_row(items[i]).strip() for i, selected in enumerate(selections) if selected and i not in hidden_columns]
            # if full_values:
            #     process = subprocess.Popen(usrtxt, shell=True, stdin=subprocess.PIPE, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
            #     process.communicate(input='\n'.join(full_values).encode('utf-8'))
            # break
        elif key in [curses.KEY_BACKSPACE, "KEY_BACKSPACE", 263, 127]:

            if cursor == 0:
                usrtxt = usrtxt[:-1]
            else:
                usrtxt = usrtxt[:-(cursor+1)] + usrtxt[-cursor:]

        elif key in [curses.KEY_LEFT, 2]:                                       # CTRL+B
            cursor = min(len(usrtxt), cursor + 1)

        elif key in [curses.KEY_RIGHT, 6]:                                      # CTRL-F
            cursor = max(0, cursor - 1)

        elif key == curses.KEY_UP:
            cursor = max(0, cursor - 1)

        elif key in [4, 330]:                                                   # Ctrl+D, Delete
            if cursor != 0 and usrtxt != "":
                if cursor == 1:
                    usrtxt = usrtxt[:-(cursor)]
                else:
                    usrtxt = usrtxt[:-(cursor)] + usrtxt[-(cursor-1):]
                cursor = max(0, cursor - 1)
                
        elif key == 21 or key == "^U":                                          # CTRL+U
            if cursor == 0:
                usrtxt = ""
            else:
                usrtxt = usrtxt[-(cursor+1):]
            cursor = len(usrtxt)
        elif key == 11 or key == "^K":                                          # CTRL+K
            if cursor: usrtxt = usrtxt[:-cursor]
            cursor = 0

        elif key in [1, 262]:                                                          # CTRL+A (beginning)
            cursor = len(usrtxt)
            
        elif key in [5, 360]:                                                          # CTRL+E (end)
            cursor = 0
        elif key in [18]:                                                          # CTRL+E (end)
            if "*" in registers:
                if cursor == 0:
                    addtxt = registers["*"]
                    usrtxt = usrtxt[-cursor:] + registers["*"]
                else:
                    usrtxt = usrtxt[:-cursor] + registers["*"] + usrtxt[-cursor:]

        # elif key in [23,8]:                                                     # Ctrl+BACKSPACE, CTRL+W
        #     if cursor == 0: tmp = usrtxt[::-1]
        #     else: tmp = usrtxt[:-cursor][::-1]
        #     index = tmp.find(" ")
        #     if index == -1: index = len(usrtxt)-1-cursor
        #
        #     cursor = len(usrtxt)
        elif key == curses.KEY_RESIZE: pass

        else:
            if isinstance(key, int):
                try:
                    val = chr(key) if chr(key).isprintable() else ''
                except:
                    val = ''
            else: val = key
            if cursor == 0:
                usrtxt += val
            else:
                usrtxt = usrtxt[:-cursor] + val + usrtxt[-cursor:]



#!/bin/python
# -*- coding: utf-8 -*-
"""
sorting.py

Author: GrimAndGreedy
License: MIT
"""

import re
from datetime import datetime
from typing import Tuple

def parse_numerical(value: str) -> float:
    """ Match first number in string and return it as a float. If not number then return INF. """
    try:
        match = re.search(r'(\d+(\.\d+)?)', value)
        if match:
            return float(match.group(1))
        return float('inf')  # Default for non-numerical values
    except ValueError:
        return float('inf')

def parse_size(value: str) -> float:
    """ Match size in string and return it as a float. If no match then return INF."""
    size_units = {
        'B': 1,
        'KB': 1024,
        'MB': 1024**2,
        'GB': 1024**3,
        'TB': 1024**4,
        'PB': 1024**5,
        'K': 1024,
        'M': 1024**2,
        'G': 1024**3,
        'T': 1024**4,
        'P': 1024**5
    }
    
    match = re.search(r'(\d+(\.\d+)?)(\s*([KMGTPEB][B]?)\s*)', value, re.IGNORECASE)
    
    if match:
        number = float(match.group(1))
        unit = match.group(4).upper() if match.group(4) else 'B'
        unit = re.sub(r'[^\w]', '', unit)  # Remove non-alphanumeric characters
        return number * size_units.get(unit, 1)
    return float('inf')  # Default for non-size values

def time_to_seconds(time_str: str) -> float:
    """Convert a time string to total seconds."""
    if time_str.strip().upper() == "INF":
        return float('inf')  # Assign infinity for "INF"

    time_units = {
        'year': 365 * 24 * 3600,
        'years': 365 * 24 * 3600,
        'day': 24 * 3600,
        'days': 24 * 3600,
        'hour': 3600,
        'hours': 3600,
        'minute': 60,
        'minutes': 60,
        'sec': 1,
        'secs': 1,
        's': 1,
        'min': 60,
    }
    
    total_seconds = 0
    try:
        tokens = time_str.split()
        for i in range(0, len(tokens), 2):
            if i + 1 < len(tokens):
                num = int(tokens[i])
                unit = tokens[i + 1]
                total_seconds += num * time_units.get(unit, 0)
    except:
        return float('inf')


    return total_seconds

def time_sort(time_str: str) -> datetime:
    """ If there is a date in the string then convert it to strptime. If no match then return 00:00 (as datetime)."""
    formats = [
        "%Y-%m-%d %H:%M",     # "2021-03-16 15:30"
        "%Y-%m-%d",           # "2021-03-16"
        "%Y/%m/%d",           # "2021/03/16"
        "%d/%m/%Y",           # "05/03/2024"
        "%A %d %b %Y %H:%M:%S",  # "Saturday 01 Feb 2025 21:19:47"
        "%a %d %b %Y %H:%M:%S",  # "Sat 01 Feb 2025 21:19:47"
        "%d/%m/%Y",           # "10/12/2023"
        "%d/%m/%y",            # "1/1/23"
        "%H:%M",               # "04:30"
        "%H:%M:%S",               # "04:30:23"
    ]
    
    for fmt in formats:
        try:
            return datetime.strptime(time_str, fmt)
        except ValueError:
            pass
    
    return datetime.strptime("00:00", "%H:%M")

def sort_items(indexed_items: list[Tuple[int,list[str]]], sort_method:int=0, sort_column:int=0, sort_reverse:bool=False):
    """ Sort indexed_items based on the sort_method on sort_column. """
    SORT_METHODS = ['original', 'lexical', 'LEXICAL', 'alphanum', 'ALPHANUM', 'time', 'numerical', 'size']
    if sort_column is not None:
        try:
            if SORT_METHODS[sort_method] == 'numerical':
                indexed_items.sort(key=lambda x: parse_numerical(x[1][sort_column]), reverse=sort_reverse)
            elif SORT_METHODS[sort_method] == 'original':
                indexed_items.sort(key=lambda x: x[0], reverse=sort_reverse)
            elif SORT_METHODS[sort_method] == 'lexical':
                # indexed_items.sort(key=lambda x: x[1][sort_column].lower(), reverse=sort_reverse)
                indexed_items.sort(key=lambda x: (1 if x[1][sort_column].strip() == "" else 0, x[1][sort_column].lower()), reverse=sort_reverse)
            elif SORT_METHODS[sort_method] == 'LEXICAL':
                indexed_items.sort(key=lambda x: (1 if x[1][sort_column].strip() == "" else 0, x[1][sort_column]), reverse=sort_reverse)
            elif SORT_METHODS[sort_method] == 'alphanum':
                indexed_items.sort(key=lambda x: (1 if x[1][sort_column].strip() == "" else 0, "".join([chr(ord('z')+ord(c)) if not c.isalnum() else c.lower() for c in x[1][sort_column]])))
            elif SORT_METHODS[sort_method] == 'time':
                indexed_items.sort(key=lambda x:time_sort(x[1][sort_column]))
            elif SORT_METHODS[sort_method] == 'ALPHANUM':
                indexed_items.sort(key=lambda x: (1 if x[1][sort_column].strip() == "" else 0, "".join([chr(ord('z')+ord(c)) if not c.isalnum() else c for c in x[1][sort_column]])))
            elif SORT_METHODS[sort_method] == 'size':
                indexed_items.sort(key=lambda x: parse_size(x[1][sort_column]), reverse=sort_reverse)
        except IndexError:
            pass  # Handle cases where sort_column is out of range
